package lexSimp;

import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import edu.northwestern.at.utils.corpuslinguistics.postagger.PartOfSpeechTagger;
import edu.northwestern.at.utils.corpuslinguistics.postagger.hepple.HeppleTagger;

import ob.core.CentralFactory;
import ob.nlp.Word;
import ob.nlp.WordScores;
import ob.nlp.WordSimilarity;
import ob.nlp.morphology.Morphology;

/**
 * @author Or Biran
 *
 */
public class RulesCreator {
	
	private static final String VERBS_FILE = "data/verbs.txt";

	private Collection<String> _verbs = Utils.readLines(new File(VERBS_FILE));
	private Morphology _morphology = CentralFactory.getMorphology();
	private PartOfSpeechTagger _tagger = new HeppleTagger();
	private Map<Word, Word> _originalForms = new HashMap<Word, Word>();

	private WordSimilarity _similarity;
	private WordScores _complexity;
	
	
	public RulesCreator(WordSimilarity similarity, WordScores complexity) {
		_similarity = similarity;
		_complexity = complexity;
	}

	public Map<Word, Map<Double, Word>> createRules() {
		Map<Word, Map<Double, Word>> rules = new TreeMap<Word, Map<Double, Word>>();
		
		for (WordSimilarity.Similarity sim : _similarity.getSimilarities()) {
			if (specialWord(sim.getWord1()) || specialWord(sim.getWord2())) continue;
			if (_complexity.getScore(sim.getWord1()) == null) continue; 
			if (_complexity.getScore(sim.getWord2()) == null) continue; 
			if (_complexity.getScore(sim.getWord1()) > _complexity.getScore(sim.getWord2())) {
				Map<Word, Word> forms = getForms(sim.getWord1(), sim.getWord2());
				for (Word word1form : forms.keySet()) {
					Map<Double, Word> wordRules = rules.get(word1form);
					if (wordRules==null) wordRules = new TreeMap<Double, Word>();
					Word word2form = forms.get(word1form);
					if (word2form.equals(word1form)) continue; // TODO: HAD TO ADD THIS...
					wordRules.put(sim.getValue(), word2form);
					rules.put(word1form, wordRules);
				}
			}
		}
		
		return rules;
	}

	public Double getComplexity(Word word) {
		Double comp = _complexity.getScore(word);
		if (comp == null) {
			Word originalForm = _originalForms.get(word);
			if (originalForm != null) {
				comp = _complexity.getScore(originalForm);
			}
		}
		if (comp == null) {
			comp = Double.NaN;
		}
		return comp;
	}

	private Map<Word, Word> getForms(Word word1, Word word2) {
		Map<Word, Word> forms = new HashMap<Word, Word>();
		
		boolean found = false;
		
		if (isVerb(word1) && isVerb(word2)) {
			found = true;
			
			String lemma1 = _morphology.lemmatize(word1.getText());
			String lemma2 = _morphology.lemmatize(word2.getText());
			
			if (lemma1 != null && lemma2 != null) {
				System.out.print(lemma1 + ", " + lemma2 + ": ");
				for (String tense : _morphology.getVerbTenses()) {
					for (String person : _morphology.getPersons()) {
						String form1 = _morphology.conjugate(lemma1, tense, person);
						String form2 = _morphology.conjugate(lemma2, tense, person);

						System.out.print("(" + form1 + "," + form2 + ") ");

						Word newWord1 = CentralFactory.getWord(form1, null);
						Word newWord2 = CentralFactory.getWord(form2, null);

						if (! forms.containsKey(newWord1)) {
							forms.put(newWord1, newWord2);
						}
					}
				}
				System.out.println();
			}
			else {
				System.out.println("null lemmas for " + word1 + ", " + word2);
			}
		}
		else if (isNoun(word1) && isNoun(word2)) {
			found = true;
			
			String lemma1 = getSingular(word1.getText());
			String lemma2 = getSingular(word2.getText());
			
			if (lemma1 != null && lemma2 != null) {

				for (int i=1; i<=2; i++) {
					String form1 = _morphology.pluralize(lemma1, i);
					String form2 = _morphology.pluralize(lemma2, i);
					Word newWord1 = CentralFactory.getWord(form1, null);
					Word newWord2 = CentralFactory.getWord(form2, null);
					if (! forms.containsKey(newWord1)) {
						forms.put(newWord1, newWord2);
					}
				}
			}
		}
		
		if (! found) forms.put(word1, word2);
		
		for (Word form1 : forms.keySet()) {
			_originalForms.put(form1, word1);
		}
		for (Word form2 : forms.values()) {
			_originalForms.put(form2, word2);
		}
		
		return forms;
	}

	private static String getSingular(String word) {
		if (word.length() < 4) return word;
		if (word.endsWith("s") && ! word.endsWith("ss")) {
			if (word.endsWith("ies")) {
				return word.substring(0, word.length() - 3) + "y";
			}
			else if (word.endsWith("ves")) {
				return word.substring(0, word.length() - 3) + "f";
			}
			else if (word.endsWith("us")) {
				return word;
			}
			else {
				return word.substring(0, word.length() - 1);
			}
		}
		return word;
	}
	
	private boolean isVerb(Word word) {
		if (word.getPos() == null) {
			if (_verbs.contains(word.getText())) return true;
			List<String> tags = _tagger.getTagsForWord(word.getText());
			for (String tag : tags) {
				if (tag.contains("v")) return true;
			}
			return false;
		}
		else return word.getPos().isVerb();
	}

	private boolean isNoun(Word word) {
		if (word.getPos() == null) {
			List<String> tags = _tagger.getTagsForWord(word.getText());
			for (String tag : tags) {
				if (tag.contains("n")) return true;
			}
			return false;
		}
		else return word.getPos().isNoun();
	}

	private static boolean specialWord(Word word) {
		return word.getText().contains("#N");
	}

}
