package lexSimp;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeSet;

import ob.core.CentralFactory;
import ob.core.Feature;
import ob.core.FeatureVector;
import ob.core.StringFeature;
import ob.nlp.SimpleWordsToTerms;
import ob.nlp.Word;
import ob.nlp.WordScores;
import ob.nlp.WordSimilarity;
import ob.nlp.corpus.ContextVectors;
import ob.nlp.util.NlpUtils;
import ob.nlp.wordNet.Synset;
import ob.nlp.wordNet.WordNet;
import ob.nlp.wordNet.jwnl.JwnlWordNet;
import ob.operators.CosineSimilarity;
import ob.util.IOUtils;

/**
 * @author Or Biran
 *
 */
public class SimplifierCreator {

	private static final String TEMP_VECTORS_PREFIX = "vectors";
	private static final String SIMILARITY_PREFIX = "similarity";
	private static final String FINAL_VECTORS_PREFIX = "all_vectors";
	private static final String COMPLEXITY_PREFIX = "complexity";
	private static final String RULES_PREFIX = "rules";
	private static final String RULES_VECTORS_PREFIX = "rules_vectors";
	private static final String SIMPLE_CORPUS = "simple";
	private static final String COMPLEX_CORPUS = "complex";
	private static final String FILTERED_INFIX = ".filtered";
	private static final String FREQUENCY_INFIX = "_counts";
	private static final String FILE_SUFFIX = ".txt";
	
	private static final int BATCH_SIZE = 1000;
	private static final int RARE_THRESHOLD = 10;
	private static final int LONG_THRESHOLD = 30;
	
	private static final String TEMPORARY_VECTORS_SUBDIR = "temp";

	private double _cutoff = 0.4;
	private int _window = 10;
	
	private String _corpusDir;
	private String _workingDir;
	
	
	public SimplifierCreator(String corpusDir, String workingDir) {
		_corpusDir = corpusDir;
		_workingDir = workingDir;
	}
	
	public void setCutoff(double cutoff) {
		_cutoff = cutoff;
	}

	public void setWindow(int window) {
		_window = window;
	}


	public void outputVectors() {
		ContextVectors vectors = CentralFactory.getContextVectors();
		
		File[] files = new File(getComplexCorpusDir()).listFiles();
		
		int lineCount = 0;
		int wordCount = 0;
		int fileCount = 0;
		
		for (File file : files) {
			BufferedReader reader = Utils.getBufferedFileReader(file);

			List<Word> words = new ArrayList<Word>();
			
			while (Utils.ready(reader)) {
				String line = Utils.readLine(reader);
				words.addAll(extractWords(line));
				lineCount++;
				
				if (words.size() > _window * 5000) {
					for (int i=_window; i<words.size() - _window; i++) {
						Word word = words.get(i);
						FeatureVector vector = vectors.getVector(word);
						if (vector == null) vector = CentralFactory.getFeatureVector();
						for (int j=i-_window; j<i+_window; j++) {
							if (i==j) continue;
							Word contextWord = words.get(j);
							if (StopWords.isStopWord(contextWord)) continue;
							vector.increment(feature(contextWord.getText()), 1.0);
						}
						vectors.setVector(word, vector);
						wordCount++;
						
						System.out.println(lineCount + " lines, " + wordCount + " words.");
					}
					
					List<Word> newWords = new ArrayList<Word>();
					for (int i=words.size() - (_window * 2); i<words.size(); i++) {
						newWords.add(words.get(i));
					}
					words = newWords;
					
					System.gc();
					Utils.printUsedMemory();
					
					if (lineCount >= 20000) {
						printFeatures(vectors, ++fileCount);
						lineCount = 0;
						wordCount = 0;
						vectors = CentralFactory.getContextVectors();;
					}
				}
				
			}
			
			Utils.close(reader);
		}
	}

	private void printFeatures(ContextVectors cv, int num) {
		String fileName = new File(getTemporaryVectorsSubdir(), getTemporaryVectorsFileName(num)).getAbsolutePath();
		IOUtils.writeContextVectors(fileName, cv);
		
		System.out.println("wrote " + num);
	}

	private String getTemporaryVectorsSubdir() {
		File subdir = new File(_workingDir, TEMPORARY_VECTORS_SUBDIR);
		subdir.mkdirs();
		return subdir.getAbsolutePath();
	}

	private static Feature feature(String name) {
		return new StringFeature(name);
	}

	private static List<Word> extractWords(String text) {
		List<String> wordStrs = NlpUtils.getWords(text);
		List<Word> words = new ArrayList<Word>();
		for (String str : wordStrs) {
			String normalized = NlpUtils.normalizeLetters(str.toLowerCase());//NlpUtils.stem(wordStr);
			words.add(CentralFactory.getWord(normalized, null));
		}
		return new SimpleWordsToTerms().wordsToTerms(words);
	}

	public void consolidateVectors() {
		Set<File> filesLeft = getFiles();
		
		Set<Word> done = new HashSet<Word>();
		ContextVectors current = CentralFactory.getContextVectors();

		Set<File> files = getFiles();
		
		for (File file : files) {
			int skipped = 0;
			filesLeft.remove(file);
			System.out.println("working on: " + file.getName());
			
			BufferedReader reader = Utils.getBufferedFileReader(file);
			
			while (Utils.ready(reader)) {
				String line = Utils.readLine(reader);
				String[] wordAndFeatures = line.split("\\s");
				Word word = CentralFactory.getWord(wordAndFeatures[0], null);
				
				if (done.contains(word)) {
					skipped++;
					continue;
				}
				
				if (current.getVector(word) == null) {
					FeatureVector vector = makeVector(Arrays.copyOfRange(wordAndFeatures, 1, wordAndFeatures.length));
					current.setVector(word, vector);
				}
				
				if (current.size() < BATCH_SIZE && Utils.ready(reader)) continue;
				
				System.out.println("skipped: " + skipped);
				
				for (File secondFile : filesLeft) {
					BufferedReader secondReader = Utils.getBufferedFileReader(secondFile);
					System.out.println("second: " + secondFile.getName());
					
					while (Utils.ready(secondReader)) {
						String secondLine = Utils.readLine(secondReader);
						String[] secondWordAndFeatures = secondLine.split("\\s");
						Word secondWord = CentralFactory.getWord(secondWordAndFeatures[0], null);
						
						if (current.getVector(secondWord) == null) continue;
						if (done.contains(secondWord)) continue;

						FeatureVector vector = makeVector(Arrays.copyOfRange(secondWordAndFeatures, 1, secondWordAndFeatures.length));
						current.getVector(secondWord).consolidate(vector);
					}
					
					Utils.close(secondReader);
				}
				
				System.out.println("smoothing..");
				
				current = smooth(current);
				
				System.out.println("saving..");
				
				String fileName = new File(_workingDir, getVectorsFileName()).getAbsolutePath();
				IOUtils.appendContextVectors(current, fileName);
				
				for (Word doneWord : current.getWords()) {
					done.add(doneWord);
				}
				
				current = CentralFactory.getContextVectors();
			}
			
			Utils.close(reader);
		}
	}

	private Set<File> getFiles() {
		File dir = new File(getTemporaryVectorsSubdir());		
		Set<File> files = new TreeSet<File>();
		for (File file : dir.listFiles()) {
			if (! file.getName().startsWith(TEMP_VECTORS_PREFIX)) continue;
//			int num = Integer.parseInt(file.getName().split("\\.")[2]);
//			if (num > 100) continue;
			files.add(file);
		}
		return files;
	}

	private static ContextVectors smooth(ContextVectors vectors) {
		ContextVectors smoothed = CentralFactory.getContextVectors();
		for (Word word : vectors.getWords()) {
			FeatureVector vector = vectors.getVector(word);
			FeatureVector smoothedVector = CentralFactory.getFeatureVector();
			for (Feature feature : vector.getFeatures()) {
				Double value = vector.getValue(feature);
				if (value > 1.0) {
					smoothedVector.set(feature, value);
				}
			}
			smoothed.setVector(word, smoothedVector);
		}
		return smoothed;
	}

	private static FeatureVector makeVector(String[] wordScores) {
		FeatureVector vector = CentralFactory.getFeatureVector();
		for (String wordScore : wordScores) {
			String[] wordAndScore = wordScore.split("::+");
			if (wordAndScore.length != 2) continue;
			String word = wordAndScore[0];
			if (word.contains("#N#")) continue;
			Double score = Double.parseDouble(wordAndScore[1]);
			vector.increment(new StringFeature(word), score);
		}
		return vector;
	}

	public void wordNetVectorFilter() {
		WordNet wordNet = JwnlWordNet.getInstance();
		
		Map<Word, Collection<Synset>> wordToSynsets = new HashMap<Word, Collection<Synset>>();
		
		BufferedReader reader = Utils.getBufferedFileReader(new File(_workingDir, getVectorsFileName()));

		ContextVectors vectors = CentralFactory.getContextVectors();

		int i = 0;
		int j = 0;
		while (Utils.ready(reader)) {
			String line = Utils.readLine(reader);
			String[] wordAndFeatures = line.split("\\s");
			i++;
			
			Word word = CentralFactory.getWord(wordAndFeatures[0], null);
			if (tooLong(word)) continue;
			
			FeatureVector vector = makeVector(Arrays.copyOfRange(wordAndFeatures, 1, wordAndFeatures.length));
			if (tooRare(vector)) continue;

			Collection<Synset> synsets = getSynsets(word, wordToSynsets, wordNet);
			if (synsets == null || synsets.isEmpty()) continue;
			
			vectors.setVector(word, vector);
			
			j++;
			
			if (i % BATCH_SIZE == 0) {
				String fileName = new File(_workingDir, getFilteredVectorsFileName()).getAbsolutePath();
				IOUtils.appendContextVectors(vectors, fileName);
				vectors = CentralFactory.getContextVectors();
				System.out.println("filtered " + j + " from " + i);
			}
		}
		
		Utils.close(reader);
	}

	private static boolean tooRare(FeatureVector vector) {
		return vector.size() < RARE_THRESHOLD;
	}

	private static boolean tooLong(Word word) {
		return word.getText().length() > LONG_THRESHOLD;
	}

	private static Collection<Synset> getSynsets(Word word, Map<Word, Collection<Synset>> wordToSynsets, WordNet wordNet) {
		Collection<Synset> synsets = wordToSynsets.get(word);
		if (synsets==null) {
			synsets = wordNet.getSynsets(word.getText(), word.getPos());
			wordToSynsets.put(word, synsets);
		}
		return synsets;
	}

	public void outputSimilarity() {
		WordSimilarity similarity = CentralFactory.getWordSimilarity();

		BufferedReader reader = Utils.getBufferedFileReader(new File(_workingDir, getFilteredVectorsFileName()));

		ContextVectors vectors = CentralFactory.getContextVectors();
		Collection<Word> seen = new HashSet<Word>();

		while (Utils.ready(reader)) {
			String line = Utils.readLine(reader);
			
			String[] wordAndFeatures = line.split("\\s");
			
			Word word = CentralFactory.getWord(wordAndFeatures[0], null);
			FeatureVector vector = makeVector(Arrays.copyOfRange(wordAndFeatures, 1, wordAndFeatures.length));
		
			vectors.setVector(word, vector);
			seen.add(word);
			
			if (vectors.size() % 10000 == 0 || ! Utils.ready(reader)) {
				similarity.consolidate(getSimilarity(vectors));
				System.out.println("merged intra similarity");
				Utils.printUsedMemory();
				System.gc();
				Utils.printUsedMemory();
				
				BufferedReader reader2 = Utils.getBufferedFileReader(new File(_workingDir, getFilteredVectorsFileName()));
				int j = 0;
				while (Utils.ready(reader2)) {
					String line2 = Utils.readLine(reader2);
					
					String[] wordAndFeatures2 = line2.split("\\s");
					Word word2 = CentralFactory.getWord(wordAndFeatures2[0], null);
					if (seen.contains(word2)) continue;
					
					FeatureVector vector1 = makeVector(Arrays.copyOfRange(wordAndFeatures2, 1, wordAndFeatures2.length));
					for (Word origWord : vectors.getWords()) {
						FeatureVector vector2 = vectors.getVector(origWord);
						Double cosSim = getCosineSimilarity(vector1, vector2);
						if (cosSim < _cutoff) continue;
						similarity.set(word2, origWord, cosSim);
					}
					
					if (++j % BATCH_SIZE == 0) {
						System.out.println(word2 + ": " + j);
						Utils.printUsedMemory();
						
						printSimilarityAppend(similarity);
						similarity = CentralFactory.getWordSimilarity();
					}
				}
				
				Utils.close(reader2);
				
				printSimilarityAppend(similarity);
				similarity = CentralFactory.getWordSimilarity();
				vectors = CentralFactory.getContextVectors();
				
				System.out.println("done with " + j);
				Utils.printUsedMemory();
				System.gc();
				Utils.printUsedMemory();
			}
		}
		
		Utils.close(reader);
	}
	
	private void printSimilarityAppend(WordSimilarity similarity) {
		String fileName = new File(_workingDir, getSimilarityFileName()).getAbsolutePath();
		IOUtils.appendWordSimilarity(fileName, similarity);
	}

	private WordSimilarity getSimilarity(ContextVectors cv) {
		WordSimilarity similarity = CentralFactory.getWordSimilarity();

		int i = 0;
		for (Word word1 : cv.getWords()) {
			i++;
			int j = 0;
			for (Word word2 : cv.getWords()) {
				if (similarity.get(word2, word1) != null) continue;
				if (word1.equals(word2)) continue;
				double cosSim = getCosineSimilarity(cv.getVector(word1), cv.getVector(word2));
				if (cosSim < _cutoff) continue;
				similarity.set(word1, word2, cosSim);
				j++;
			}
			System.out.println(i + ". " + word1 + ", " + j);
			Utils.printUsedMemory();
		}
		
		return similarity;
	}

	private static CosineSimilarity _cosineSimilarity = new CosineSimilarity();
	
	private static double getCosineSimilarity(FeatureVector vec1, FeatureVector vec2) {
		return _cosineSimilarity.getValue(vec1, vec2);
	}

	public void wordNetSimilarityFilter() {
		
		String filename = new File(_workingDir, getSimilarityFileName()).getAbsolutePath();
		
		System.out.println("reading..");
		
		WordSimilarity similarity = IOUtils.loadSimilarities(filename);
		
		System.out.println("loaded similarity");
		
		WordNet wordNet = JwnlWordNet.getInstance();
		
		System.out.println("loaded dictionary");

		WordSimilarity filtered = CentralFactory.getWordSimilarity();
		
		System.out.println("filtering similarity..");
		
		Map<Word, Collection<Synset>> wordToSynsets = new HashMap<Word, Collection<Synset>>();
		
		int i = 0;
		int j = 0;
		for (Word word1 : similarity.getFirstWords()) {
			i++;
			System.out.println("word " + i + " of " + similarity.size() + ". filtered: " + j);
			for (Word word2 : similarity.getSecondWords(word1)) {
				j += findReplacableSimilarity(word1, word2, wordToSynsets, wordNet, similarity, filtered);
				j += findReplacableSimilarity(word2, word1, wordToSynsets, wordNet, similarity, filtered);
			}
			
			if (i % BATCH_SIZE == 0) {
				System.out.println("writing..");
				String outputFile = new File(_workingDir, getFilteredSimilarityFileName()).getAbsolutePath();
				IOUtils.appendWordSimilarity(outputFile, filtered);
				filtered = CentralFactory.getWordSimilarity();
			}
		}
		
		System.out.println("writing..");
		String outputFile = new File(_workingDir, getFilteredSimilarityFileName()).getAbsolutePath();
		IOUtils.appendWordSimilarity(outputFile, filtered);
	}

	private static int findReplacableSimilarity(Word word1, Word word2,
			Map<Word, Collection<Synset>> wordToSynsets, WordNet wordNet,
			WordSimilarity similarity, WordSimilarity filtered) {

		Collection<Synset> synsets = getSynsets(word1, wordToSynsets, wordNet);
		Synset synset = chooseSense(synsets);
		
		if (synset != null && isReplacable(synset, word2)) {
			Double currentSimilarity = filtered.get(word1, word2);
			Double candidateSimilarity = similarity.get(word1, word2);
			if (candidateSimilarity != null) {
				if (currentSimilarity == null || currentSimilarity < candidateSimilarity) {
					filtered.set(word1, word2, candidateSimilarity);
					return 1;
				}
			}
		}
		
		return 0;
	}

	private static Synset chooseSense(Collection<Synset> synsets) {
		return synsets.isEmpty() ? null : synsets.iterator().next();
	}

	private static boolean isReplacable(Synset synset, Word otherWord) {
		// synonyms
		if (Utils.containsIgnoreCase(synset.getSynonyms(), otherWord.getText())) {
			return true;
		}

		// hypernyms
		for (Synset hypernym : synset.getHypernyms()) {
			if (Utils.containsIgnoreCase(hypernym.getSynonyms(), otherWord.getText())) {
				return true;
			}
		}
		
		return false;
	}

	public void outputComplexWordFrequency() {
		File[] files = new File(getComplexCorpusDir()).listFiles();

		outputWordFrequency(files, new File(_workingDir, getComplexFrequencyFileName()));
	}

	public void outputSimpleWordFrequency() {
		File[] files = new File(getSimpleCorpusDir()).listFiles();

		outputWordFrequency(files, new File(_workingDir, getSimpleFrequencyFileName()));
	}

	private void outputWordFrequency(File[] files, File outputFile) {
		WordScores counts = CentralFactory.getWordScores();
		double normalizer = 0;
		
		for (File file : files) {
			System.out.println("reading " + file);
			BufferedReader reader = Utils.getBufferedFileReader(file);

			int i = 0;
			
			while (Utils.ready(reader)) {
				String line = Utils.readLine(reader);
				List<Word> words = extractWords(line);
				for (Word word : words) {
					counts.increment(word, 1.0);
					normalizer++;
				}
				
				if (++i % 10000 == 0) {
					System.out.println(i);
					Utils.printUsedMemory();
				}
			}
			
			Utils.close(reader);
		}

		BufferedWriter writer = Utils.getBufferedFileWriter(outputFile);
		for (Word word : counts.getWords()) {
			if (StopWords.isStopWord(word)) continue;
			double normalScore = counts.getScore(word) / normalizer;
			Utils.write(writer, word.getText() + "\t" + normalScore + "\n");
		}
		Utils.close(writer);
	}
	
	public void outputComplexityScores() {
		String simpleFrequencyFile = new File(_workingDir, getSimpleFrequencyFileName()).getAbsolutePath();
		String complexFrequencyFile = new File(_workingDir, getComplexFrequencyFileName()).getAbsolutePath();
		
		WordScores simpleFrequency = IOUtils.readWordScores(simpleFrequencyFile);
		WordScores complexFrequency = IOUtils.readWordScores(complexFrequencyFile);
		
		double simpleUnseen = min(simpleFrequency);
		double complexUnseen = min(complexFrequency);
		
		WordScores complexityScores = CentralFactory.getWordScores();
		
		for (Word word : simpleFrequency.getWords()) {
			double sFreq = simpleFrequency.getScore(word);
			double cFreq = Utils.coalesce(complexFrequency.getScore(word), complexUnseen);

			if (sFreq != simpleUnseen && cFreq != complexUnseen) {
				double complexity = (cFreq * complexFrequency.getWords().size()) / (sFreq * simpleFrequency.getWords().size());
				complexityScores.set(word, complexity);
			}
		}

		for (Word word : complexFrequency.getWords()) {
			if (complexityScores.getWords().contains(word)) continue;
			
			double sFreq = Utils.coalesce(simpleFrequency.getScore(word), simpleUnseen);
			double cFreq = complexFrequency.getScore(word);

			if (sFreq != simpleUnseen && cFreq != complexUnseen) {
				double complexity = (cFreq * complexFrequency.getWords().size()) / (sFreq * simpleFrequency.getWords().size());
				complexityScores.set(word, complexity);
			}
		}

		for (Word word : new TreeSet<Word>(complexityScores.getWords())) {
			double wordComplexity = getWordComplexity(word);
			double baseComplexity = complexityScores.getScore(word);
			Double frequency = complexFrequency.getScore(word);
			if (frequency==null) frequency = complexUnseen;
			frequency *= complexFrequency.getWords().size();
			System.out.println(word + "(" + Utils.formatDouble(baseComplexity * wordComplexity) + "): " + Utils.formatDouble(wordComplexity) + ", " + Utils.formatDouble(baseComplexity));
			complexityScores.set(word, baseComplexity * wordComplexity * (1 / frequency));
		}
		
		String outputFile = new File(_workingDir, getComplexityFileName()).getAbsolutePath();
		
		IOUtils.writeWordScores(outputFile, complexityScores);
	}

	private static String getComplexityFileName() {
		return COMPLEXITY_PREFIX + FILE_SUFFIX;
	}

	private static double getWordComplexity(Word word) {
		return (double)word.getText().length() / 7.0;
	}

	private static double min(WordScores ws) {
		double min = Double.POSITIVE_INFINITY;
		for (Word word : ws.getWords()) min = Math.min(ws.getScore(word), min);
		return min;
	}

	public void outputRules() {
		String similarityFile = new File(_workingDir, getFilteredSimilarityFileName()).getAbsolutePath();
		String complexityFile = new File(_workingDir, getComplexityFileName()).getAbsolutePath();

		WordSimilarity similarity = IOUtils.loadSimilarities(similarityFile);
		WordScores complexity = IOUtils.readWordScores(complexityFile);
		
		RulesCreator rulesCreator = new RulesCreator(similarity, complexity);

		Map<Word, Map<Double, Word>> rules = rulesCreator.createRules();
		
		String outputFile = new File(_workingDir, getRulesFileName()).getAbsolutePath();
		IOUtils.writeRules(outputFile, rules);
		
		for (Word word : rules.keySet()) {
			Double comp1 = rulesCreator.getComplexity(word);
			System.out.println(word + " (" + Utils.formatDouble(comp1) + "): ");
			Map<Double, Word> wordRules = rules.get(word);
			for (Double score : wordRules.keySet()) {
				Word toWord = wordRules.get(score);
				Double comp2 = rulesCreator.getComplexity(toWord);
				System.out.println("-> " + toWord + " (" + Utils.formatDouble(comp2) + "): " + score);
			}
		}
		
		System.out.println(rules.size() + " rules created");
	}

	public void outputRulesVectors() {
		String rulesFile = new File(_workingDir, getRulesFileName()).getAbsolutePath();
		Map<Word, SortedMap<Double, Word>> rules = IOUtils.readRules(rulesFile);

		Collection<Word> ruleWords = new HashSet<Word>();
		
		for (Word word1 : rules.keySet()) {
			ruleWords.add(word1);
			ruleWords.addAll(rules.get(word1).values());
		}
		
		File[] files = new File(getComplexCorpusDir()).listFiles();

		ContextVectors vectors = CentralFactory.getContextVectors();

		int lineCount = 0;
		int wordCount = 0;
		int tempWordCount = 0;
		boolean done = false;

		for (File file : files) {
			System.out.println("reading " + file);
			if (done) break;
			BufferedReader reader = Utils.getBufferedFileReader(file);

			List<Word> words = new ArrayList<Word>();
			
			while (!done && Utils.ready(reader)) {
				String line = Utils.readLine(reader);
				words.addAll(extractWords(line));
				lineCount++;
				
				if (words.size() > _window * 5000) {
					for (int i=_window; i<words.size() - _window; i++) {
						Word word = words.get(i);
						if (! ruleWords.contains(word)) continue;
						
						FeatureVector vector = vectors.getVector(word);
						if (vector == null) vector = CentralFactory.getFeatureVector();
						for (int j=i-_window; j<i+_window; j++) {
							if (i==j) continue;
							Word contextWord = words.get(j);
							if (StopWords.isStopWord(contextWord)) continue;
							if (contextWord.getText().endsWith(":")) continue; // had to add this to avoid breaking the vectors reader
							vector.increment(feature(contextWord.getText()), 1.0);
						}
						vectors.setVector(word, vector);
						wordCount++;
						tempWordCount++;
						
						System.out.println(lineCount + " lines, " + wordCount + " words.");
					}
					
					List<Word> newWords = new ArrayList<Word>();
					for (int i=words.size() - (_window * 2); i<words.size(); i++) {
						newWords.add(words.get(i));
					}
					words = newWords;
					
					if (tempWordCount > 1000000) {
						vectors = smooth(vectors);
						tempWordCount = 0;
					}
					
					Utils.printUsedMemory();
				}
			}
			
			Utils.close(reader);
		}
		
		String fileName = new File(_workingDir, getRulesVectorsFileName()).getAbsolutePath();
		IOUtils.writeContextVectors(fileName, vectors);
	}

	private String getRulesVectorsFileName() {
		return getRulesVectorsFileName(getFeaturesName());
	}

	private String getTemporaryVectorsFileName(int num) {
		return TEMP_VECTORS_PREFIX + getFeaturesName() + FILE_SUFFIX + "." + num;
	}

	private String getVectorsFileName() {
		return FINAL_VECTORS_PREFIX + getFeaturesName() + FILE_SUFFIX;
	}

	private String getFilteredVectorsFileName() {
		return FINAL_VECTORS_PREFIX + getFeaturesName() + FILTERED_INFIX + FILE_SUFFIX;
	}

	private String getSimilarityFileName() {
		return SIMILARITY_PREFIX + getFeaturesName() + FILE_SUFFIX;
	}

	private String getFilteredSimilarityFileName() {
		return SIMILARITY_PREFIX + getFeaturesName() + FILTERED_INFIX + FILE_SUFFIX;
	}

	private String getRulesFileName() {
		return getRulesFileName(getFeaturesName());
	}

	private String getFeaturesName() {
		return getFeatureName(_window);
	}
	
	public static String getFeatureName(int window) {
		return "_" + "n" + window;
	}

	public static String getComplexFrequencyFileName() {
		return COMPLEX_CORPUS + FREQUENCY_INFIX + FILE_SUFFIX;
	}

	public static String getSimpleFrequencyFileName() {
		return SIMPLE_CORPUS + FREQUENCY_INFIX + FILE_SUFFIX;
	}

	private String getComplexCorpusDir() {
		return new File(_corpusDir, COMPLEX_CORPUS).getAbsolutePath();
	}

	private String getSimpleCorpusDir() {
		return new File(_corpusDir, SIMPLE_CORPUS).getAbsolutePath();
	}

	public static String getRulesFileName(String featuresName) {
		return RULES_PREFIX + featuresName + FILE_SUFFIX;
	}

	public static String getRulesVectorsFileName(String featuresName) {
		return RULES_VECTORS_PREFIX + featuresName + FILE_SUFFIX;
	}
}
