package lexSimp;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import ob.core.CentralFactory;
import ob.core.Feature;
import ob.core.FeatureVector;
import ob.core.NumericFeature;
import ob.nlp.SimpleWordExtractor;
import ob.nlp.SimpleWordsToTerms;
import ob.nlp.Word;
import ob.operators.CosineSimilarity;

/**
 * @author Or Biran
 *
 */
public class Utils {


	public static void printUsedMemory() {
		long free = Runtime.getRuntime().freeMemory();
		long total = Runtime.getRuntime().totalMemory();
		long used = total - free;
		System.out.println("Used memory: " + used + " of " + total);
	}


	public static void close(BufferedWriter writer) {
		try {
			writer.close();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static void write(BufferedWriter writer, String string) {
		try {
			writer.write(string);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}


	public static void close(BufferedReader reader) {
		try {
			reader.close();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static BufferedReader getBufferedFileReader(File file) {
		try {
			return new BufferedReader(new FileReader(file));
		} catch (FileNotFoundException e) {
			throw new RuntimeException(e);
		}
	}

	public static BufferedWriter getBufferedFileWriter(File file) {
		try {
			return new BufferedWriter(new FileWriter(file));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static String readLine(BufferedReader reader) {
		try {
			return reader.readLine();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static boolean ready(BufferedReader reader) {
		try {
			return reader.ready();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static List<String> readLines(File file) {
		List<String> lines = new ArrayList<String>();
		BufferedReader reader = getBufferedFileReader(file);
		while (ready(reader)) {
			lines.add(readLine(reader));
		}
		close(reader);
		return lines;
	}

	public static String formatDouble(Double d) {
		return NumberFormat.getInstance().format(d);
	}

	public static <T> T coalesce(T value, T defaultValue) {
		return value == null ? defaultValue : value;
	}

	public static boolean containsIgnoreCase(Collection<String> strings, String target) {
		for (String string : strings) {
			if (string.toLowerCase().equals(target.toLowerCase())) return true;
		}
		return false;
	}

	public static FeatureVector getCommonFeatures(FeatureVector vector1, FeatureVector vector2) {
		FeatureVector commonVector = CentralFactory.getFeatureVector();
		for (Feature feature : vector1.getFeatures()) {
			if (vector2.getFeatures().contains(feature)) {
				double value = Math.min(vector1.getValue(feature), vector2.getValue(feature));
				commonVector.set(feature, value);
			}
		}
		return commonVector;
	}
	
	public static FeatureVector makeContextVector(List<Word> context, int index, String featuresName) {
		FeatureVector vector = CentralFactory.getFeatureVector();
		
		String[] featureNames = featuresName.split("_");
		
		for (String featureName : featureNames) {
			
			if (featureName.startsWith("n")) {
				int window = Integer.parseInt(featureName.substring(1));
				int start = Math.max(0, index - window);
				int end = Math.min(context.size(), index + window + 1);
				for (int i = start; i<end; i++) {
					if (i==index) continue;
					Word word = context.get(i);
					if (StopWords.isStopWord(word)) continue;
					Feature feature = new NumericFeature(word.getText()/* + "_" + window + "n"*/);
					vector.increment(feature, 1.0);
				}
			}
			
			if (featureName.startsWith("a")) {
				int a = Integer.parseInt(featureName.substring(1));
				int end = Math.min(context.size() - 1, index + a);
				for (int i = index+1; i < end; i++) {
					Word word = context.get(i);
					Feature feature = new NumericFeature(word + "_0b" + a + "a");
					vector.increment(feature, 1.0);
				}
			}

			if (featureName.startsWith("b")) {
				int b = Integer.parseInt(featureName.substring(1));
				int start = Math.max(0, index - b);
				for (int i = start; i < index; i++) {
					Word word = context.get(i);
					Feature feature = new NumericFeature(word + "_" + b + "b0a");
					vector.increment(feature, 1.0);
				}
			}
			
		}
		
		return vector;
	}

	
	public static boolean equalContextWord(String contextWord, String word) {
		if (contextWord.equalsIgnoreCase(word)) return true;
		if (contextWord.contains("#N#") && hasDigits(word)) return true;
		return false;
	}

	public static boolean hasDigits(String word) {
		for (char c : word.toCharArray()) if (Character.isDigit(c)) return true;
		return false;
	}

	public static List<Word> getContextWords(String line) {
		List<Word> words = new SimpleWordExtractor().extractWords(line);
		return new SimpleWordsToTerms().wordsToTerms(words);
	}

	public static double similarityInContext(FeatureVector candidateVector, FeatureVector wordVector, List<Word> context, int index, String featuresName) {
		FeatureVector commonVector = Utils.getCommonFeatures(candidateVector, wordVector);
		FeatureVector contextVector = Utils.makeContextVector(context, index, featuresName);
		
//		printVector(commonVector);
//		printVector(contextVector);
		
		double similarity = new CosineSimilarity().getValue(commonVector, contextVector);

		return similarity;
	}

//	private static void printVector(FeatureVector vector) {
//		for (Feature feature : vector.getFeatures()) {
//			System.out.print(feature.getName() + ", ");
//		}
//		System.out.println();
//	}


}
