package lexSimp.app;

import java.io.BufferedReader;
import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;

import lexSimp.SimplifierCreator;
import lexSimp.Utils;

import edu.northwestern.at.utils.corpuslinguistics.postagger.PartOfSpeechTagger;
import edu.northwestern.at.utils.corpuslinguistics.postagger.hepple.HeppleTagger;

import ob.core.CentralFactory;
import ob.core.FeatureVector;
import ob.nlp.Word;
import ob.nlp.WordScores;
import ob.nlp.corpus.ContextVectors;
import ob.nlp.morphology.Morphology;
import ob.nlp.util.NlpUtils;
import ob.nlp.wordNet.Synset;
import ob.nlp.wordNet.WordNet;
import ob.nlp.wordNet.jwnl.JwnlWordNet;
import ob.operators.CosineSimilarity;
import ob.util.IOUtils;

public class EvaluationRunner {

	private static final double SIMILARITY_THRESHOLD = 0.01;
	private static final double CONTEXTUALITY_THRESHOLD = 0.1;
	private static final int MIN_SENTENCE_LENGTH = 7;

	private static final int WINDOW = 10;
	private static final boolean VERBOSE = true;
	
	private static final String VERBS_FILE = "data/verbs.txt";
	
	private static String _complexFrequencyFile;

	
	public static void main(String[] args) throws Exception {
		if (args.length != 2) {
			System.out.println("USAGE: EvaluationRunner <rulesDir> <inputFile>");
			System.exit(0);
		}
		
		String workingDir = args[0]; //"working/output/wikipedia";
		String inputFile = args[1]; //"working/output/toy/input_large.txt";
		
		_complexFrequencyFile = new File(workingDir, SimplifierCreator.getComplexFrequencyFileName()).getAbsolutePath();
		
		String rulesFile = new File(workingDir, SimplifierCreator.getRulesFileName(getFeaturesName())).getAbsolutePath();
		Map<Word, SortedMap<Double, Word>> rules = IOUtils.readRules(rulesFile);
		
		String vectorsFile = new File(workingDir, SimplifierCreator.getRulesVectorsFileName(getFeaturesName())).getAbsolutePath();
		ContextVectors vectors = IOUtils.readContextVectors(vectorsFile);
		
		BufferedReader reader = Utils.getBufferedFileReader(new File(inputFile));

		Map<String, Collection<String>> alreadySimplified = new HashMap<String, Collection<String>>();
		
		int c = 0;
		
		int identical = 0;		// sentences which were simplified identically by the system and baseline
		int multiple = 0;		// sentences in which more than one word was simplified by the system
		int repeated = 0;		// sentences in which a single word was simplified, but that word was already simplified in an earlier sentence
		int repeatedWord = 0;	// sentences in which a single word was simplified, but the sentence already contains the replacement word
		
		List<String> items = new ArrayList<String>();
		
		while (Utils.ready(reader)) {
			String line = Utils.readLine(reader);
			List<String> sentences = NlpUtils.splitSentences(line);
			
			for (String sentence : sentences) {
				Simplified simplified = simplify(sentence, rules, vectors);
				if (simplified.getIndeces().size() > 1) {
					multiple++;
					continue;
				}
				
				String simplifiedWord = simplified.getSimplifiedWord();
				if (simplifiedWord == null) continue;
				
				String baseline = simplifyBaseline(simplified);
				
				if (baseline.equals(simplified.getSimplified())) {
					identical++;
					continue;
				}
				
				if (Utils.containsIgnoreCase(simplified.getWords(), simplifiedWord)) {
					repeatedWord++;
					continue;
				}

				String originalWord = simplified.getOriginalWord();
				
				Collection<String> alreadySimplifiedWords = alreadySimplified.get(originalWord);
				if (alreadySimplifiedWords == null) alreadySimplifiedWords = new ArrayList<String>();
				
				if (alreadySimplifiedWords.contains(simplifiedWord)) {
					repeated++;
					continue;
				}
				
				alreadySimplifiedWords.add(simplifiedWord);
				alreadySimplified.put(originalWord, alreadySimplifiedWords);
				
				double frequency = getFrequencies().getScore(CentralFactory.getWord(originalWord, null));
				String item = frequency + "\t" + simplified.getSimilarity() + "\t" + sentence + "\t" + simplified.getSimplified() + "\t" + baseline;
				
				items.add(item);
				
				if (items.size() % 1 == 0) {
					printItems(items);
					
					if (VERBOSE) {
						System.out.println(++c);

						System.out.println("identical: " + identical);
						System.out.println("multiple: " + multiple);
						System.out.println("repeated: " + repeated);
					}

					items = new ArrayList<String>();
				}
			}
		}
		
		printItems(items);
		
		if (VERBOSE) {
			System.out.println("identical: " + identical);
			System.out.println("multiple: " + multiple);
			System.out.println("repeated: " + repeated);
		}
		
		Utils.close(reader);
	}
	
	private static String getFeaturesName() {
		return SimplifierCreator.getFeatureName(WINDOW);
	}

	private static void printItems(List<String> items) {
		for (String item : items) {
			System.out.println(item);
		}
	}

	private static Morphology _morphology = CentralFactory.getMorphology();
	private static PartOfSpeechTagger _tagger = new HeppleTagger();

	private static WordNet _wordNet = JwnlWordNet.getInstance();
	private static Collection<String> _verbs = Utils.readLines(new File(VERBS_FILE));

	private static Map<String, Collection<Synset>> _wordToSynsets = new HashMap<String, Collection<Synset>>();

	private static WordScores _frequencies;

	private static WordScores getFrequencies() {
		if (_frequencies == null) {
			_frequencies = IOUtils.readWordScores(_complexFrequencyFile);
		}
		return _frequencies;
	}

	private static String simplifyBaseline(Simplified simplified) {
		String baseline = "";
		boolean changed = false;
		for (int i=0; i<simplified.getWords().size(); i++) {
			String word = simplified.getWords().get(i);
			if (simplified.getIndeces().contains(i)) {
				String newWord = simplifyBaselineWord(word).toUpperCase();
				if (! word.equalsIgnoreCase(newWord)) {
					word = newWord;
					changed = true;
				}
			}
			baseline += word + " ";
		}
		return changed ? baseline : simplified.getOriginal();
	}

	private static String simplifyBaselineWord(String word) {
		String best = word;
		Double wordScore = getFrequencies().getScore(word(word));
		if (wordScore!=null) { 
			Double highest = wordScore;

			Collection<Synset> synsets = getSynsets(word, _wordToSynsets, _wordNet);
			Synset synset = chooseSense(synsets);

			if (synset == null) return best;
			for (String synonym : synset.getSynonyms()) {
				if (synonym.length() > 4 && word.length() > 4 && synonym.substring(0, 4).equalsIgnoreCase(word.substring(0, 4))) break;
				Double synonymScore = getFrequencies().getScore(word(synonym));
				if (synonymScore == null) continue;
				if (synonymScore > highest) {
					highest = synonymScore;
					best = synonym;
				}
			}
		}
		
		if (! best.equals(word)) {
			if (isVerb(best) && isVerb(word)) {
				String lemma = _morphology.lemmatize(word);
				String bestLemma = _morphology.lemmatize(best);
				
				for (String tense : _morphology.getVerbTenses()) {
					for (String person : _morphology.getPersons()) {
						String form = _morphology.conjugate(lemma, tense, person);
						if (form.equals(word)) {
							return _morphology.conjugate(bestLemma, tense, person);
						}
					}
				}
			}
			else if (isNoun(best) && isNoun(word)) {
				String lemma = getSingular(word);
				String bestLemma = getSingular(best);

				for (int i=1; i<=2; i++) {
					String form = _morphology.pluralize(lemma, i);
					if (form.equals(word)) {
						return _morphology.pluralize(bestLemma, i);
					}
				}
			}
		}
		
		return best;
	}

	private static String getSingular(String word) {
		if (word.length() < 4) return word;
		if (word.endsWith("s") && ! word.endsWith("ss")) {
			if (word.endsWith("ies")) {
				return word.substring(0, word.length() - 3) + "y";
			}
			else if (word.endsWith("ves")) {
				return word.substring(0, word.length() - 3) + "f";
			}
			else {
				return word.substring(0, word.length() - 1);
			}
		}
		return word;
	}
	
	private static boolean isVerb(String word) {
		if (_verbs.contains(word)) return true;
		List<String> tags = _tagger.getTagsForWord(word);
		for (String tag : tags) {
			if (tag.contains("v")) return true;
		}
		return false;
	}

	private static boolean isNoun(String word) {
		List<String> tags = _tagger.getTagsForWord(word);
		for (String tag : tags) {
			if (tag.contains("n")) return true;
		}
		return false;
	}


	private static Synset chooseSense(Collection<Synset> synsets) {
		return synsets.isEmpty() ? null : synsets.iterator().next();
	}

	private static Word word(String word) {
		return CentralFactory.getWord(word, null);
	}

	private static Collection<Synset> getSynsets(String word, Map<String, Collection<Synset>> wordToSynsets, WordNet wordNet) {
		Collection<Synset> synsets = wordToSynsets.get(word);
		if (synsets==null) {
			synsets = wordNet.getSynsets(word, null);
			wordToSynsets.put(word, synsets);
		}
		return synsets;
	}

	private static Simplified simplify(String sentence, Map<Word, SortedMap<Double, Word>> rules, ContextVectors vectors) {
		Simplified simplified = new Simplified();
		simplified.setOriginal(sentence);
		
		List<String> words = getWords(sentence);
		simplified.setWords(words);
		List<Word> contextWords = Utils.getContextWords(sentence);
		if (contextWords.size() < MIN_SENTENCE_LENGTH) return simplified;
		
		List<String> newWords = new ArrayList<String>();
		int index = 0;
		String enclosure = null;
		for (int i=0; i<words.size(); i++) {
			if (index < contextWords.size() - 1 && Utils.equalContextWord(contextWords.get(index).getText(), words.get(i))) {
				index++;
			}
			Word word = CentralFactory.getWord(words.get(i), null);
			if (enclosure != null) {
				if (isEndEnclosure(word.getText(), enclosure)) {
					enclosure = null;
				}
				newWords.add(word.getText());
			}
			else if (isEnclosure(word.getText())) {
				enclosure = word.getText();
				newWords.add(word.getText());
			}
			else if (rules.containsKey(word)) {
				Word replacement = word;
				FeatureVector wordVector = vectors.getVector(word);
				if (wordVector==null) {
					newWords.add(word.getText());
					continue;
				}
				for (Word candidate : rules.get(word).values()) {
					FeatureVector candidateVector = vectors.getVector(candidate);
					if (candidateVector==null) continue;
					double similarity = Utils.similarityInContext(candidateVector, wordVector, contextWords, index, getFeaturesName());
					
					double contextuality = getContextuality(wordVector, contextWords, index);
					
					boolean override = candidateAlreadyInSentence(sentence, candidate.getText());
					if (contextuality > CONTEXTUALITY_THRESHOLD) override = true;
					
					if (similarity > SIMILARITY_THRESHOLD && ! override) {
						simplified.setSimilarity(similarity);
						replacement = candidate;
						simplified.addIndex(i);
					}
				}
				newWords.add(replacement.getText());
			}
			else newWords.add(word.getText());
		}

		String newText = "";
		for (int i=0; i<newWords.size(); i++) {
			String word = newWords.get(i);
			if (simplified.getIndeces().contains(i)) {
				word = word.toUpperCase();
			}
			newText += word + " ";
		}
		simplified.setSimplified(newText);
		simplified.setNewWords(newWords);
		
		return simplified;
	}
	
	private static List<String> getWords(String line) {
		return NlpUtils.getWords(line);
	}
	
	private static final String[] ENCLOSURES = {
		"-LRB-",
		"``"
	};
	private static final String[] END_ENCLOSURES = {
		"-RRB-",
		"''"
	};
	
	private static boolean isEnclosure(String word) {
		for (String enclosure : ENCLOSURES) {
			if (word.equals(enclosure)) return true;
		}
		return false;
	}

	private static boolean isEndEnclosure(String word, String enclosure) {
		for (int i=0; i<ENCLOSURES.length; i++) {
			if (enclosure.equals(ENCLOSURES[i])) {
				if (word.equals(END_ENCLOSURES[i])) return true;
				else return false;
			}
		}
		throw new RuntimeException("something horrible happened");
	}

	private static boolean candidateAlreadyInSentence(String sentence, String candidate) {
		if (sentence.toLowerCase().contains(candidate)) return true;
		if (sentence.toLowerCase().contains(_morphology.lemmatize(candidate))) return true;
		if (sentence.toLowerCase().contains(getSingular(candidate))) return true;
		return false;
	}

	private static double getContextuality(FeatureVector wordVector, List<Word> contextWords, int index) {
		FeatureVector contextVector = Utils.makeContextVector(contextWords, index, getFeaturesName());
		return new CosineSimilarity().getValue(wordVector, contextVector);
	}
	
	
	private static class Simplified {
		private String _original;
		private List<String> _words;
		private List<String> _newWords;
		private String _simplified;
		private Double _similarity;
		private Collection<Integer> _indeces = new ArrayList<Integer>();
		
		public String getOriginal() {
			return _original;
		}
		public String getOriginalWord() {
			if (_indeces.isEmpty()) return null;
			if (_words.size() <= _indeces.iterator().next()) return null;
			return _words.get(_indeces.iterator().next());
		}
		public String getSimplifiedWord() {
			if (_indeces.isEmpty()) return null;
			if (_newWords.size() <= _indeces.iterator().next()) return null;
			return _newWords.get(_indeces.iterator().next());
		}
		public void setOriginal(String original) {
			_original = original;
		}
		public String getSimplified() {
			return _simplified;
		}
		public void setSimplified(String simplified) {
			_simplified = simplified;
		}
		public Collection<Integer> getIndeces() {
			return _indeces;
		}
		public void addIndex(Integer index) {
			_indeces.add(index);
		}
		public List<String> getWords() {
			return _words;
		}
		public void setWords(List<String> words) {
			_words = words;
		}
		public void setNewWords(List<String> newWords) {
			_newWords = newWords;
		}
		public Double getSimilarity() {
			return _similarity;
		}
		public void setSimilarity(Double similarity) {
			_similarity = similarity;
		} 
		
	}
}
