package ob.nlp.corpus;

import java.util.List;

import ob.core.CentralFactory;
import ob.core.Feature;
import ob.core.FeatureVector;
import ob.core.StringFeature;
import ob.nlp.StopWords;
import ob.nlp.StopWordsFromFiles;
import ob.nlp.Word;


/**
 * @author Or Biran
 *
 * Builds context vectors from counts of occurances of other words as the k near 
 * neighbors of a word.
 */
public class NeighborContextVectorsBuilder implements ContextVectorsBuilder {

	private int _neighborsBefore;
	private int _neighborsAfter;
	private StopWords _stopWords = new StopWordsFromFiles();
	private boolean _useContextStopWords = false;
	
	public NeighborContextVectorsBuilder(int numNeighbors) {
		this(numNeighbors, numNeighbors);
	}

	public NeighborContextVectorsBuilder(int neighborsBefore, int neighborsAfter) {
		_neighborsBefore = neighborsBefore;
		_neighborsAfter = neighborsAfter;
	}

	@Override
	public ContextVectors buildVectors(List<Word> words) {
		ContextVectors vectors = CentralFactory.getContextVectors();
		
		for (int i=0; i<words.size(); i++) {
			Word word = words.get(i);
			if (isStopWord(word)) continue;
			FeatureVector vector = CentralFactory.getFeatureVector();
			int beginIndex = Math.max(i - _neighborsBefore, 0);
			int endIndex = Math.min(i + _neighborsAfter + 1, words.size());
			for (int j=beginIndex; j<endIndex; j++) {
				if (i==j) continue;
				Word contextWord = words.get(j);
				if (!isUseContextStopWords() && isStopWord(contextWord)) continue;
				vector.increment(feature(name(contextWord), contextWord), 1.0);
			}
			
			vectors.setVector(word, vector);
		}
		
		return vectors;
	}

	private Feature feature(String name, Word word) {
		return new StringFeature(name);
//		return new WordFeature(name, word);
	}

	private String name(Word word) {
		if (_neighborsBefore == _neighborsAfter) {
			return wordName(word) + "_" + _neighborsBefore + "n";
		}
		else {
			return wordName(word) + "_" + _neighborsBefore + "b" + _neighborsAfter + "a";
		}
	}

	private String wordName(Word word) {
//		return word.getText() + "_" + pos(word);
		return word.getText();
	}

//	private String pos(Word word) {
//		return word.getPos() == null ? "untagged" : word.getPos().getString();
//	}

	private boolean isStopWord(Word word) {
		return _stopWords.isStopWord(word);
	}

	public StopWords getStopWords() {
		return _stopWords;
	}

	public void setStopWords(StopWords stopWords) {
		_stopWords = stopWords;
	}

	public boolean isUseContextStopWords() {
		return _useContextStopWords;
	}

	public void setUseContextStopWords(boolean useContextStopWords) {
		_useContextStopWords = useContextStopWords;
	}

	
}
