package ob.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import ob.core.CentralFactory;
import ob.core.Feature;
import ob.core.FeatureVector;
import ob.core.NumericFeature;
import ob.nlp.Word;
import ob.nlp.WordScores;
import ob.nlp.WordSimilarity;
import ob.nlp.WordSimilarity.Similarity;
import ob.nlp.corpus.ContextVectors;

/**
 * @author Or Biran
 * 
 * Provides various Input/Output utility methods
 */
public class IOUtils {

	private static final String FEATURE_SCORE_SEPERATOR = "::";
	private static final String TAB_DELIM = "\t";
	private static final String EOL = "\n";

	public static WordSimilarity loadSimilarities(String filename) {
		WordSimilarity similarity = CentralFactory.getWordSimilarity();
		
		BufferedReader reader = getBufferedReader(filename);
		
		while (ready(reader)) {
			String[] tokens = readLine(reader).split(TAB_DELIM);
			assert (tokens.length == 3);
			String word1 = tokens[0];
			String word2 = tokens[1];
			String simStr = tokens[2];
			Double sim = Double.parseDouble(simStr);
			similarity.set(parseWord(word1), parseWord(word2), sim);
		}
		
		return similarity;
	}

	private static BufferedReader getBufferedReader(String filename) {
		try {
			return new BufferedReader(new FileReader(filename));
		} catch (FileNotFoundException e) {
			throw new RuntimeException(e);
		}
	}

	private static String readLine(BufferedReader reader) {
		try {
			return reader.readLine();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static boolean ready(BufferedReader reader) {
		try {
			return reader.ready();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static WordScores readWordScores(String file) {
		WordScores scores = CentralFactory.getWordScores();
		for (String line : Utils.readLines(file)) {
			String[] items = line.split(TAB_DELIM);
			scores.set(parseWord(items[0]), Double.parseDouble(items[1]));
		}
		return scores;
	}

	public static void writeWordSimilarity(String file, WordSimilarity wordSimilarity) {
		BufferedWriter writer = getBufferedWriter(file);
		for (WordSimilarity.Similarity sim : wordSimilarity.getSimilarities()) {
			write(writer, similarityToString(sim) + EOL);
		}
		close(writer);
	}

	private static BufferedWriter getBufferedWriter(String file) {
		try {
			return new BufferedWriter(new FileWriter(file));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static BufferedWriter getAppendBufferedWriter(String file) {
		try {
			return new BufferedWriter(new FileWriter(file, true));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static void write(BufferedWriter writer, String string) {
		try {
			writer.write(string);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static void close(BufferedWriter writer) {
		try {
			writer.close();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private static String similarityToString(Similarity sim) {
		return writeWord(sim.getWord1()) + TAB_DELIM + writeWord(sim.getWord2()) + TAB_DELIM + sim.getValue();
	}
	
	private static String writeWord(Word word) {
		return word.getText();
	}

	private static Word parseWord(String text) {
		return CentralFactory.getWord(text, null);
	}


	public static void writeWordScores(String file, WordScores scores) {
		BufferedWriter writer = getBufferedWriter(file);
		for (Word word : scores.getWords()) {
			write(writer, word.getText() + TAB_DELIM + scores.getScore(word) + EOL);
		}
		close(writer);
	}

	public static void writeRules(String file, Map<Word, Map<Double, Word>> rules) {
		BufferedWriter writer = getBufferedWriter(file);
		for (Word word : rules.keySet()) {
			write(writer, writeWord(word));
			Map<Double, Word> wordRules = rules.get(word);
			for (Double d : wordRules.keySet()) {
				write(writer, TAB_DELIM + wordRules.get(d) + TAB_DELIM + d);
			}
			write(writer, EOL);
		}
		close(writer);
	}

	public static Map<Word, SortedMap<Double, Word>> readRules(String file) {
		Map<Word, SortedMap<Double, Word>> rules = new TreeMap<Word, SortedMap<Double, Word>>();
		for (String line : Utils.readLines(file)) {
			String[] items = line.split(TAB_DELIM);
			Word word1 = parseWord(items[0]);
			SortedMap<Double, Word> wordRule = new TreeMap<Double, Word>();
			for (int i=1; i<items.length; i+=2) {
				Word word2 = parseWord(items[i]);
				Double d = Double.parseDouble(items[i+1]);
				wordRule.put(d, word2);
			}
			rules.put(word1, wordRule);
		}
		return rules;
	}

	public static void writeContextVectors(String fileName, ContextVectors cv) {
		BufferedWriter writer = getBufferedWriter(fileName);
		
		for (Word word : cv.getWords()) {
			FeatureVector vector = cv.getVector(word);
			String line = word.toString();
			for (Feature feature : vector.getSortedFeatures()) {
				line += TAB_DELIM + feature.getName() + FEATURE_SCORE_SEPERATOR + vector.getValue(feature);
			}
			write(writer, line + EOL);
		}

		close(writer);
	}

	public static ContextVectors readContextVectors(String fileName) {
		ContextVectors vectors = CentralFactory.getContextVectors();
		
		for (String line : Utils.readLines(fileName)) {
			String[] items = line.split(TAB_DELIM);
			Word word = parseWord(items[0]);
			FeatureVector vector = CentralFactory.getFeatureVector();
			for (int i=1; i<items.length; i++) {
				String[] featureAndScore = items[i].split(FEATURE_SCORE_SEPERATOR); 
				Feature feature = new NumericFeature(featureAndScore[0]);
				double value = Double.parseDouble(featureAndScore[1]);
				vector.set(feature, value);
			}
			vectors.setVector(word, vector);
		}
		
		return vectors;
	}

	public static void appendContextVectors(ContextVectors cv, String fileName) {
		BufferedWriter writer = getAppendBufferedWriter(fileName);
		
		for (Word word : cv.getWords()) {
			FeatureVector vector = cv.getVector(word);
			String line = word.toString();
			int skip = vector.getFeatures().size() - 100;
			for (Feature feature : vector.getSortedFeatures()) {
				if (skip-- > 0) continue;
				line += TAB_DELIM + feature.getName() + FEATURE_SCORE_SEPERATOR + vector.getValue(feature);
			}
			write(writer, line + EOL);
		}

		close(writer);
		
	}

	public static void appendWordSimilarity(String file, WordSimilarity wordSimilarity) {
		BufferedWriter writer = getAppendBufferedWriter(file);
		for (WordSimilarity.Similarity sim : wordSimilarity.getSimilarities()) {
			write(writer, similarityToString(sim) + EOL);
		}
		close(writer);
	}

	public static void appendLinesToFile(String file, List<String> lines) {
		BufferedWriter writer = getAppendBufferedWriter(file);
		for (String line : lines) {
			write(writer, line + EOL);
		}
		close(writer);
	}

}
